#include "CommonUtils/FileCommentRemover.hpp"
#include "CommonUtils/TabInterpolator.hpp"
#include "CommonUtils/TabAdjuster.hpp"
#include "CommonUtils/DigitAdjuster.hpp"
#include "NuclearDataUtils/OtherUtils/Endf6FileStore.hpp"
#include "ContinuousEnergyXSUtils/AceDataParser/FastAceDataParser.hpp"
#include "ContinuousEnergyXSUtils/AceDataParser/ThermalScatterAceDataParser.hpp"
#include "TabUtils/IntegrateTabData.hpp"
#include "TabUtils/LinearizeTabData.hpp"

using namespace frendy;

////////////////////////////////////////////////////////////////////////////////////////////////////

void read_nu_data(FastAceDataObject& ace_obj, int mt_no, vector<Real>& e_vec, vector<Real>& nu_vec)
{
  ace_obj.get_xs_data(1, e_vec, nu_vec);
  nu_vec.clear();

  vector<Real> nu_coef;
  if( mt_no == 1452 )
  {
    nu_coef = ace_obj.get_nu_data();
  }
  else if( mt_no == 1456 )
  {
    nu_coef = ace_obj.get_nu_p_data();
  }
  else if( mt_no == 1455 )
  {
    nu_coef = ace_obj.get_nu_d_data();
  }
  else
  {
    return;
  }

  int i_max = static_cast<int>(e_vec.size());
  int j_max = static_cast<int>(nu_coef.size());
  if( j_max == 0 )
  {
    e_vec.clear();
    nu_vec.clear();
    return;
  }

  nu_vec.resize(i_max);
  for(int i=0; i<i_max; i++)
  {
    Real e_coef = 1.0;
    nu_vec[i]   = 0.0;
    for(int j=0; j<j_max; j++)
    {
      nu_vec[i] += nu_coef[j] * e_coef;
      e_coef    *= e_vec[i];
    }
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void read_ace_file_fast(string file_name, vector<int>& mt_list,
                        vector<vector<Integer> >& nbt_xs,   vector<vector<Integer> >& int_xs,
                        vector<vector<Real> >&    e_int_xs, vector<vector<Real> >&    xs_tab)
{
  //Read ACE file
  FastAceDataParser read_obj;
  read_obj.set_file_name(file_name);
  FastAceDataObject ace_obj = read_obj.get_ace_data_obj();

  int i_max = static_cast<int>(mt_list.size());
  if( i_max == 0 )
  {
    mt_list.push_back(  1); //total
    mt_list.push_back(  2); //scatter;
    mt_list.push_back(101); //disappearance

    int fis_flg = 0;
    vector<int> mt_list_add = ace_obj.get_react_type_list();
    for(int i=0; i<static_cast<int>(mt_list_add.size()); i++)
    {
      mt_list.push_back(mt_list_add[i]);
      if( mt_list_add[i] == 18 )
      {
        fis_flg = 1;
      }
    }
    mt_list_add.clear();

    if( fis_flg > 0 )
    {
      //For chi (fission spectrum)
      mt_list.push_back(1018);

      //For nu_tot
      if( static_cast<int>(ace_obj.get_nu_data().size()) > 0 )
      {
        int mt_add = static_cast<int>(ace_obj.get_nu_bar_repr_flg())*1000 + 452; 
        mt_list.push_back(mt_add);
      }

      //For nu_f
      if( static_cast<int>(ace_obj.get_nu_d_data().size()) > 0 &&
          static_cast<int>(ace_obj.get_nu_d_bar_repr_flg()) == 2 )
      {
        int mt_add = static_cast<int>(ace_obj.get_nu_d_bar_repr_flg())*1000 + 455; 
        mt_list.push_back(mt_add);
      }

      //For nu_p
      if( static_cast<int>(ace_obj.get_nu_p_data().size()) > 0 )
      {
        int mt_add = static_cast<int>(ace_obj.get_nu_p_bar_repr_flg())*1000 + 456; 
        mt_list.push_back(mt_add);
      }
    }
    
    i_max   = static_cast<int>(mt_list.size());
  }

  //Resize xs data
  nbt_xs.resize(i_max);
  int_xs.resize(i_max);
  e_int_xs.resize(i_max);
  xs_tab.resize(i_max);
  
  for(int i=0; i<i_max; i++)
  {
    if( mt_list[i] < 1000 )
    {
      ace_obj.get_xs_data(mt_list[i], e_int_xs[i], xs_tab[i]);
    }
    else
    {
      if( mt_list[i] == 1018 )
      {
        ace_obj.get_chi_data(e_int_xs[i], xs_tab[i]);
      }
      else if( mt_list[i] < 2000 )
      {
        read_nu_data(ace_obj, mt_list[i], e_int_xs[i], xs_tab[i]);
      }
      else if( mt_list[i] == 2452 )
      {
        e_int_xs[i] = ace_obj.get_nu_ene_data();
        xs_tab[i]   = ace_obj.get_nu_data();
      }
      else if( mt_list[i] == 2455 )
      {
        e_int_xs[i] = ace_obj.get_nu_d_ene_data();
        xs_tab[i]   = ace_obj.get_nu_d_data();
      }
      else if( mt_list[i] == 2456 )
      {
        e_int_xs[i] = ace_obj.get_nu_p_ene_data();
        xs_tab[i]   = ace_obj.get_nu_p_data();
      }
    }

    int j_max = static_cast<int>(e_int_xs[i].size());
    for(int j=0; j<j_max; j++)
    {
      e_int_xs[i][j] *= 1.0E6; //MeV -> eV
    }
    nbt_xs[i].resize(1);
    int_xs[i].resize(1);
    nbt_xs[i][0] = j_max;
    int_xs[i][0] = int_lin_lin;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void read_ace_file_tsl(string file_name, vector<int>& mt_list,
                       vector<vector<Integer> >& nbt_xs,   vector<vector<Integer> >& int_xs,
                       vector<vector<Real> >&    e_int_xs, vector<vector<Real> >&    xs_tab)
{
  //Read ACE file
  ThermalScatterAceDataParser read_obj;
  read_obj.set_file_name(file_name);
  ThermalScatterAceDataObject ace_obj = read_obj.get_ace_data_obj();

  vector<Real> coh_ene   = ace_obj.get_coh_elastic_ene_data();
  vector<Real> coh_xs    = ace_obj.get_coh_elastic_e_xs_data();
  vector<Real> incoh_ene = ace_obj.get_incoh_elastic_ene_data();
  vector<Real> incoh_xs  = ace_obj.get_incoh_elastic_xs_data();
  vector<Real> inela_ene = ace_obj.get_inelastic_ene_data();
  vector<Real> inela_xs  = ace_obj.get_inelastic_xs_data();

  int i_max = static_cast<int>(mt_list.size());
  if( i_max == 0 )
  {
    if( static_cast<int>(inela_xs.size()) > 0 )
    {
      mt_list.push_back(221); //inelastic scattering
    }

    if( static_cast<int>(coh_xs.size()) > 0 )
    {
      mt_list.push_back(222); //coherent elastic scattering
    }

    if( static_cast<int>(incoh_xs.size()) > 0 )
    {
      mt_list.push_back(223); //incoherent elastic scattering
    }

    i_max = static_cast<int>(mt_list.size());
  }

  //Resize xs data
  nbt_xs.resize(i_max);
  int_xs.resize(i_max);
  e_int_xs.resize(i_max);
  xs_tab.resize(i_max);

  for(int i=0; i<i_max; i++)
  {
    if( mt_list[i] == 221 )
    {
      e_int_xs[i] = inela_ene;
      xs_tab[i]   = inela_xs;
    }

    if( mt_list[i] == 222 )
    {
      e_int_xs[i] = coh_ene;
      xs_tab[i]   = coh_xs;
    }

    if( mt_list[i] == 223 )
    {
      e_int_xs[i] = incoh_ene;
      xs_tab[i]   = incoh_xs;
    }

    int j_max = static_cast<int>(e_int_xs[i].size());
    for(int j=0; j<j_max; j++)
    {
      e_int_xs[i][j] *= 1.0E6; //MeV -> eV
    }
    nbt_xs[i].resize(1);
    int_xs[i].resize(1);
    nbt_xs[i][0] = j_max;
    int_xs[i][0] = int_lin_lin;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void read_weight_data(string weight_file_name, vector<Real> e_int_weight, vector<Real>& weight_tab)
{
  FileCommentRemover file_mod_obj;
  vector<string> input_vec = file_mod_obj.remove_comment(weight_file_name);
  file_mod_obj.clear();

  StringUtils   str_obj;
  VectorClearer clr_obj;
  str_obj.set_text_data(input_vec);
  clr_obj.clear_vec_array1_str(input_vec);

  vector<vector<string> > input_data = str_obj.get_div_string_vec_all();
  str_obj.clear();

  vector<Real> e_weight_ori, weight_tab_ori;
  e_weight_ori.clear();
  weight_tab_ori.clear();

  int i_max  = static_cast<int>(input_data.size());
  int ele_no = 0;
  for(int i=0; i<i_max; i++)
  {
    int j_max = static_cast<int>(input_data[i].size());
    for(int j=0; j<j_max; j++)
    {
      if( static_cast<int>(input_data[i][j].length()) > 0 )
      {
        if( ele_no%2 == 0 )
        {
          e_weight_ori.push_back(numeric_string_cast<Real>(input_data[i][j]));
        }
        else //if( ele_no%2 == 1 )
        {
          weight_tab_ori.push_back(numeric_string_cast<Real>(input_data[i][j]));
        }
        ele_no++;
      }
    }
  }
  if( ele_no %2 == 1 )
  {
    int pos_no = static_cast<int>(weight_tab_ori.size());
    weight_tab_ori.push_back(weight_tab_ori[pos_no-1]);
  }

  vector<Integer> nbt_vec, int_vec;
  nbt_vec.resize(1);
  int_vec.resize(1);
  nbt_vec[0] = static_cast<int>(e_weight_ori.size());
  int_vec[0] = int_lin_lin;

  TabInterpolator  ti_obj;

  int ene_no = static_cast<int>(e_int_weight.size()); 
  weight_tab.resize(ene_no);
  for(int i=0; i<ene_no; i++)
  {
    weight_tab[i] = 0.0;
    ti_obj.interpolation_tab1(e_int_weight[i], weight_tab[i],
                              nbt_vec, int_vec, e_weight_ori, weight_tab_ori);
  }
  nbt_vec.clear();
  int_vec.clear();
  e_weight_ori.clear();
  weight_tab_ori.clear();
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void set_weight(string weight_mode, vector<Real>& e_vec_ori, vector<Real>& weight_tab)
{
  int ene_no = static_cast<int>(e_vec_ori.size());

  weight_tab.resize(ene_no);
  if( weight_mode == "const" )
  {
    for(int i=0; i<ene_no; i++)
    {
      weight_tab[i] = 1.0;
    }
  }
  else if( weight_mode == "1/e" )
  {
    for(int i=0; i<ene_no; i++)
    {
      weight_tab[i] = 1.0 / e_vec_ori[i];
    }
  }
  else
  {
    read_weight_data(weight_mode, e_vec_ori, weight_tab);
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void multi_group_generator(string mg_file_name, string weight_mode,
                           vector<vector<Integer> >& nbt_xs,   vector<vector<Integer> >& int_xs,
                           vector<vector<Real> >&    e_int_xs, vector<vector<Real> >&    xs_tab)
{
  ifstream fin(mg_file_name.c_str());
  vector<Real> mg_grid, mg_el, mg_eh, mg_ref;
  
  if(!fin)
  {
    cout << endl;
    cout << "Multi-group structure data file : " << mg_file_name << endl;
    cout << "Multi-group structure data file cannot be found." << endl;
    throw runtime_error("Please check multi-group structure data file.");
  }

  while(!fin.eof())
  {
    Real grid_data;
    fin >> grid_data;
    mg_grid.push_back(grid_data);
  }

  TabAdjuster ad_obj;
  ad_obj.delete_overlap_grid(mg_grid);

  cout.precision(7);
  cout.setf(ios::showpoint);
  cout.setf(ios::scientific);
  //cout << "    Energy grid structure [MeV]" << endl;
  int unit_check = 0; //MeV
  for(int i=0; i<static_cast<int>(mg_grid.size()); i++)
  {
    if( mg_grid[i] < 1.0E-10 )
    {
      break;
    }
    else if( mg_grid[i] > 1.0E+6 )
    {
      unit_check = 1; //eV
      break;
    }
  }
  if( unit_check == 0 )
  {
    cout << "The unit of multigroup energy group structure is modified from MeV to eV." << endl;
    for(int i=0; i<static_cast<int>(mg_grid.size()); i++)
    {
      //cout << "      " << i+1 << "\t" << mg_grid[i] << endl;
      mg_grid[i] *= 1.0E+6; //[MeV] -> [eV]
    }
  }

  DigitAdjuster   digit_obj;
  int mg_no = static_cast<int>(mg_grid.size()) - 1;
  mg_ref.resize(mg_no*2);
  mg_el.resize(mg_no);
  mg_eh.resize(mg_no);
  for(int i=0; i<mg_no; i++)
  {
    mg_ref[2*i]   = digit_obj.get_adjusted_value(mg_grid[i],   sig_fig, -1.0); //el of each energy grid
    mg_ref[2*i+1] = digit_obj.get_adjusted_value(mg_grid[i+1], sig_fig, +1.0); //eh of each energy grid
  }

  TabInterpolator  ti_obj;
  IntegrateTabData integ_obj;
  int i_max = static_cast<int>(xs_tab.size());
  for(int i=0; i<i_max; i++)
  {
    if(static_cast<int>(xs_tab[i].size()) == 0 )
    {
      nbt_xs[i].clear();
      int_xs[i].clear();
      e_int_xs[i].clear();
      xs_tab[i].clear();
      continue;
    }

    vector<Real> weight_tab;
    set_weight(weight_mode, e_int_xs[i], weight_tab);

    vector<Real> xs_tab_mod = xs_tab[i];
    int j_max = static_cast<int>(xs_tab_mod.size());
    for(int j=0; j<j_max; j++)
    {
      xs_tab_mod[j] *= weight_tab[j];
    }

    vector<Integer> nbt_weight, int_weight;
    nbt_weight.resize(1);
    int_weight.resize(1);
    nbt_weight[0] = j_max;
    int_weight[0] = int_lin_lin;

    Real ene_min = e_int_xs[i][0];
    Real ene_max = e_int_xs[i][static_cast<int>(e_int_xs[i].size())-1];

    vector<Real> xs_tab_mg;
    xs_tab_mg.resize(mg_no*2);
    for(int j=0; j<mg_no; j++)
    {
      Real xs_mg_val     = 0.0;
      Real weight_mg_val = 0.0;
      Real el            = mg_grid[j];
      Real eh            = mg_grid[j+1];
      if( ene_min > eh || ene_max < el )
      {
        xs_mg_val = 0.0;
      }
      else
      {
        if( ene_min > el )
        {
          el = ene_min;
        }

        if( ene_max < eh )
        {
          eh = ene_max;
        }
        xs_mg_val = integ_obj.integrate_tab1_data(nbt_xs[i], int_xs[i], e_int_xs[i], xs_tab_mod, el, eh);
        xs_mg_val /= (eh - el);

        weight_mg_val = integ_obj.integrate_tab1_data(nbt_weight, int_weight, e_int_xs[i], weight_tab, el, eh);
        weight_mg_val /= (eh - el);

        if( fabs(weight_mg_val) > min_value )
        {
          xs_mg_val /= weight_mg_val;
        }
      }
      xs_tab_mg[2*j]   = xs_mg_val;
      xs_tab_mg[2*j+1] = xs_mg_val;
    }

    xs_tab[i]   = xs_tab_mg;
    e_int_xs[i] = mg_ref;
    nbt_xs[i].resize(1);
    int_xs[i].resize(1);
    nbt_xs[i][0] = mg_no;
    int_xs[i][0] = int_lin_lin;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void modify_xs_data(vector<vector<Real> >& e_ref,
                    vector<vector<Integer> >& nbt_xs,   vector<vector<Integer> >& int_xs,
                    vector<vector<Real> >&    e_int_xs, vector<vector<Real> >&    xs_tab)
{
  TabInterpolator ti_obj;

  int i_max = static_cast<int>(e_ref.size());
  for(int i=0; i<i_max; i++)
  {
    vector<Integer> nbt_mod, int_mod;
    vector<Real>    xs_mod;

    if( static_cast<int>(xs_tab[i].size()) == 0 )
    {
      nbt_xs[i].clear();
      int_xs[i].clear();
      e_int_xs[i].clear();
      xs_tab[i].clear();
      continue;
    }

    int j_max = static_cast<int>(e_ref[i].size());
    nbt_mod.push_back(j_max);
    int_mod.push_back(int_lin_lin);

    xs_mod.resize(j_max);
    for(int j=0; j<j_max; j++)
    {
      ti_obj.interpolation_tab1(e_ref[i][j], xs_mod[j], nbt_xs[i], int_xs[i], e_int_xs[i], xs_tab[i]);
    }
    nbt_xs[i]   = nbt_mod;
    int_xs[i]   = int_mod;
    e_int_xs[i] = e_ref[i];
    xs_tab[i]   = xs_mod;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void output_result(string out_name, string ace_file_name_ref,      string ace_file_name_comp,
                                    string mg_structure_file_name, string mg_weight_opt,
                   vector<Integer> xs_type_list,      vector<vector<Real> >& ene_vec,
                   vector<vector<Real> >& xs_ref_vec, vector<vector<Real> >& xs_comp_vec)
{
  int xs_data_no = static_cast<int>(xs_type_list.size());
  for(int i=0; i<xs_data_no; i++)
  {
    if( static_cast<int>(ene_vec[i].size())     == 0 ||
        static_cast<int>(xs_ref_vec[i].size())  == 0 ||
        static_cast<int>(xs_comp_vec[i].size()) == 0 )
    {
      continue;
    }

    ostringstream  oss;
    oss << xs_type_list[i];

    string out_name_mod = out_name + "_mt";
    if( xs_type_list[i] < 10 )
    {
      out_name_mod = out_name_mod + "00";
    }
    else if( xs_type_list[i] < 100 )
    {
      out_name_mod = out_name_mod + "0";
    }
    out_name_mod = out_name_mod + oss.str() + ".out";

    ofstream fout;
    fout.open(out_name_mod.c_str());

    fout << "# comp data             : " << ace_file_name_comp << endl;
    fout << "# ref  data             : " << ace_file_name_ref  << endl;
    if( static_cast<int>(mg_structure_file_name.size()) > 0 )
    {
      fout << "# multi group structure : " << mg_structure_file_name << endl;
      fout << "# weight mode           : " << mg_weight_opt << endl;
    }
    fout << "#" << endl;
    fout << "# energy [eV]   / comp data       / ref data        / relative dif (comp - ref)/ref" << endl;

    fout.precision(10);
    fout.setf(ios::showpoint);
    fout.setf(ios::scientific);

    int j_max = static_cast<int>(ene_vec[i].size());
    for(int j=0; j<j_max; j++)
    {
      Real xs_dif = xs_comp_vec[i][j] - xs_ref_vec[i][j];
      if( fabs(xs_ref_vec[i][j]) > min_value )
      {
        xs_dif /= xs_ref_vec[i][j];
      }

      fout << ene_vec[i][j] << "  "
           << xs_comp_vec[i][j] << "  " << xs_ref_vec[i][j] << "  " << xs_dif << endl;
    }
    fout.close();
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

int main(int argc, char *argv[])
{
  if( argc <= 1 )
  {
    throw runtime_error("Please set reference ace file name.");
  }
  else if( argc <= 2 )
  {
    throw runtime_error("Please set comparing ace file name.");
  }
  else if( argc <= 3 )
  {
    throw runtime_error("Please set output file name.");
  }

  int arg_no  =  4;
  int tsl_flg = -1;
  if( argc > arg_no )
  {
    string tsl_check = argv[arg_no];
    transform(tsl_check.begin(), tsl_check.end(), tsl_check.begin(), ::tolower);

    string fast_mode  = "fast";
    string tsl_mode01 = "thermal";
    string tsl_mode02 = "therm";
    string tsl_mode03 = "tsl";
    if( tsl_check == fast_mode )
    {
      arg_no++;
      tsl_flg = 0;
      cout << "  ACE file mode : fast" << endl;
    }
    else if( tsl_check == tsl_mode01 || tsl_check == tsl_mode02 || tsl_check == tsl_mode03 )
    {
      arg_no++;
      tsl_flg = 1;
      cout << "  ACE file mode : thermal (tsl)" << endl;
    }
  }

  string  ace_file_name_ref  = argv[1];
  string  ace_file_name_comp = argv[2];
  string  output_name        = argv[3];

  vector<int> mt_list;
  mt_list.clear();
//  mt_list.push_back(  1); //total
//  mt_list.push_back(  2); //scatter;
//  mt_list.push_back(101); //disappearance


  vector<vector<Integer> > nbt_xs_r, nbt_xs_c, int_xs_r, int_xs_c;
  vector<vector<Real> >    e_int_xs_r, e_int_xs_c, xs_tab_r, xs_tab_c;

  cout << "  read_ace_file (ref)  " << ace_file_name_ref  << endl;
  if( tsl_flg <= 0 )
  {
    read_ace_file_fast(ace_file_name_ref,  mt_list,
                       nbt_xs_r, int_xs_r, e_int_xs_r, xs_tab_r);
  }
  else
  {
    read_ace_file_tsl(ace_file_name_ref,  mt_list,
                      nbt_xs_r, int_xs_r, e_int_xs_r, xs_tab_r);
  }

  cout << "  read_ace_file (comp) " << ace_file_name_comp << endl;
  if( tsl_flg <= 0 )
  {
    read_ace_file_fast(ace_file_name_comp,  mt_list,
                       nbt_xs_c, int_xs_c, e_int_xs_c, xs_tab_c);
  }
  else
  {
    read_ace_file_tsl(ace_file_name_comp,  mt_list,
                      nbt_xs_c, int_xs_c, e_int_xs_c, xs_tab_c);
  }

  string mg_structure_file_name = "";
  string weight_mode            = "";
  if( argc > arg_no )
  {
    mg_structure_file_name = argv[arg_no];

    weight_mode = "const";
    if( argc > arg_no+1 )
    {
      string mg_weight_file_name = argv[arg_no+1];
      string mg_weight_opt       = mg_weight_file_name;
      transform(mg_weight_opt.begin(), mg_weight_opt.end(), mg_weight_opt.begin(), ::tolower);
      if( mg_weight_opt == "const" )
      {
        weight_mode = "const";
      }
      else if( mg_weight_opt == "1/e" )
      {
        weight_mode = "1/e";
      }
      else
      {
        weight_mode = mg_weight_file_name;
      }
    }

    cout << "  multi-group generation using " << mg_structure_file_name << endl;
    cout << "  weighting data                " << weight_mode           << endl;
    multi_group_generator(mg_structure_file_name, weight_mode,
                          nbt_xs_r,   int_xs_r,   e_int_xs_r,   xs_tab_r);
    multi_group_generator(mg_structure_file_name, weight_mode,
                          nbt_xs_c,   int_xs_c,   e_int_xs_c,   xs_tab_c);
  }
  else
  {
    cout << "  modify_xs_data" << endl;
    modify_xs_data(e_int_xs_r, nbt_xs_c, int_xs_c, e_int_xs_c, xs_tab_c);
  }

  //output ene, xs_r, xs_c
  cout << "  output_results" << endl;
  output_result(output_name, ace_file_name_ref,      ace_file_name_comp,
                             mg_structure_file_name, weight_mode,
                mt_list, e_int_xs_r, xs_tab_r, xs_tab_c);
}

