#include "CommonUtils/TabInterpolator.hpp"
#include "CommonUtils/TabAdjuster.hpp"
#include "CommonUtils/DigitAdjuster.hpp"
#include "NuclearDataUtils/OtherUtils/Endf6FileStore.hpp"
#include "ContinuousEnergyXSUtils/AceDataParser/ThermalScatterAceDataParser.hpp"
#include "ContinuousEnergyXSUtils/OtherUtils/IntegrateTabData.hpp"

using namespace frendy;

static const int react_no      = 3;
static const int coh_elastic   = 0;
static const int incoh_elastic = 1;
static const int inelastic     = 2;

////////////////////////////////////////////////////////////////////////////////////////////////////

void read_ace_file(string file_name, vector<vector<Real> >& e_int_xs, vector<vector<Real> >& xs_tab)
{
  //Read ACE file
  ThermalScatterAceDataParser read_obj;
  read_obj.set_file_name(file_name);
  ThermalScatterAceDataObject ace_obj = read_obj.get_ace_data_obj();

  e_int_xs.resize(react_no);
  xs_tab.resize(react_no);

  e_int_xs[coh_elastic] = ace_obj.get_coh_elastic_ene_data();
  xs_tab[coh_elastic]   = ace_obj.get_coh_elastic_e_xs_data();

  e_int_xs[incoh_elastic] = ace_obj.get_incoh_elastic_ene_data();
  xs_tab[incoh_elastic]   = ace_obj.get_incoh_elastic_xs_data();

  e_int_xs[inelastic] = ace_obj.get_inelastic_ene_data();
  xs_tab[inelastic]   = ace_obj.get_inelastic_xs_data();
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void multi_group_generator(string mg_file_name,
                           vector<vector<Integer> >& nbt_xs,   vector<vector<Integer> >& int_xs,
                           vector<vector<Real> >&    e_int_xs, vector<vector<Real> >&    xs_tab)
{
  ifstream fin(mg_file_name.c_str());
  vector<Real> mg_grid, mg_el, mg_eh, mg_ref;
  
  if(!fin)
  {
    cout << endl;
    cout << "Multi-group structure data file : " << mg_file_name << endl;
    cout << "Multi-group structure data file cannot be found." << endl;
    throw runtime_error("Please check multi-group structure data file.");
  }

  while(!fin.eof())
  {
    Real grid_data;
    fin >> grid_data;
    mg_grid.push_back(grid_data);
  }

  TabAdjuster ad_obj;
  ad_obj.delete_overlap_grid(mg_grid);

  cout.precision(7);
  cout.setf(ios::showpoint);
  cout.setf(ios::scientific);
  //cout << "    Energy grid structure [MeV]" << endl;
  for(int i=0; i<static_cast<int>(mg_grid.size()); i++)
  {
    //cout << "      " << i+1 << "\t" << mg_grid[i] << endl; 
    mg_grid[i] *= 1.0E+6; //[MeV] -> [eV]
  }

  DigitAdjuster digit_obj;
  int mg_no = static_cast<int>(mg_grid.size()) - 1;
  mg_ref.resize(mg_no*2);
  mg_el.resize(mg_no);
  mg_eh.resize(mg_no);
  for(int i=0; i<mg_no; i++)
  {
    mg_ref[2*i]   = digit_obj.get_adjusted_value(mg_grid[i],   sig_fig, -1.0); //el of each energy grid
    mg_ref[2*i+1] = digit_obj.get_adjusted_value(mg_grid[i+1], sig_fig, +1.0); //eh of each energy grid
  }

  IntegrateTabData integ_obj;
  int i_max = static_cast<int>(xs_tab.size());
  for(int i=0; i<i_max; i++)
  {
    if(static_cast<int>(xs_tab[i].size()) == 0 )
    {
      nbt_xs[i].clear();
      int_xs[i].clear();
      e_int_xs[i].clear();
      xs_tab[i].clear();
      continue;
    }

    Real ene_min = e_int_xs[i][0];
    Real ene_max = e_int_xs[i][static_cast<int>(e_int_xs[i].size())-1];

    vector<Real> xs_tab_mg;
    xs_tab_mg.resize(mg_no*2);
    for(int j=0; j<mg_no; j++)
    {
      Real el, eh, xs_mg_val;
      el = mg_grid[j];
      eh = mg_grid[j+1];
      if( ene_min > eh || ene_max < el )
      {
        xs_mg_val = 0.0;
      }
      else
      {
        if( ene_min > el )
        {
          el = ene_min;
        }

        if( ene_max < eh )
        {
          eh = ene_max;
        }
        xs_mg_val = integ_obj.integrate_tab1_data(nbt_xs[i], int_xs[i], e_int_xs[i], xs_tab[i], el, eh);
        xs_mg_val /= (eh - el);
      }
      xs_tab_mg[2*j]   = xs_mg_val;
      xs_tab_mg[2*j+1] = xs_mg_val;
    }

    xs_tab[i]   = xs_tab_mg;
    e_int_xs[i] = mg_ref;
    nbt_xs[i].resize(1);
    int_xs[i].resize(1);
    nbt_xs[i][0] = mg_no;
    int_xs[i][0] = int_lin_lin;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void modify_xs_data(vector<vector<Real> >& e_ref,
                    vector<vector<Integer> >& nbt_xs,   vector<vector<Integer> >& int_xs,
                    vector<vector<Real> >&    e_int_xs, vector<vector<Real> >&    xs_tab)
{
  TabInterpolator ti_obj;

  int i_max = static_cast<int>(e_ref.size());
  for(int i=0; i<i_max; i++)
  {
    vector<Integer> nbt_mod, int_mod;
    vector<Real>    xs_mod;

    if( static_cast<int>(xs_tab[i].size()) == 0 )
    {
      nbt_xs[i].clear();
      int_xs[i].clear();
      e_int_xs[i].clear();
      xs_tab[i].clear();
      continue;
    }

    int j_max = static_cast<int>(e_ref[i].size());
    nbt_mod.push_back(j_max);
    int_mod.push_back(int_lin_lin);

    xs_mod.resize(j_max);
    for(int j=0; j<j_max; j++)
    {
      ti_obj.interpolation_tab1(e_ref[i][j], xs_mod[j], nbt_xs[i], int_xs[i], e_int_xs[i], xs_tab[i]);
    }
    nbt_xs[i]   = nbt_mod;
    int_xs[i]   = int_mod;
    e_int_xs[i] = e_ref[i];
    xs_tab[i]   = xs_mod;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void output_result(string out_name, vector<vector<Real> >& ene_vec,
             vector<vector<Real> >& xs_ref_vec, vector<vector<Real> >& xs_comp_vec)
{
  for(int i=0; i<react_no; i++)
  {
    if( static_cast<int>(ene_vec[i].size())     == 0 ||
        static_cast<int>(xs_ref_vec[i].size())  == 0 ||
        static_cast<int>(xs_comp_vec[i].size()) == 0 )
    {
      continue;
    }

    string out_name_mod = out_name;
    if( i == coh_elastic )
    {
      out_name_mod = out_name + "_coh_ela";
    }
    else if( i == incoh_elastic )
    {
      out_name_mod = out_name + "_incoh_ela";
    }
    else if( i == inelastic )
    {
      out_name_mod = out_name + "_inela";
    }
    out_name_mod = out_name_mod + ".out";

    ofstream fout;
    fout.open(out_name_mod.c_str());

    fout.precision(10);
    fout.setf(ios::showpoint);
    fout.setf(ios::scientific);

    int j_max = static_cast<int>(ene_vec[i].size());
    for(int j=0; j<j_max; j++)
    {
      Real xs_dif = xs_comp_vec[i][j] - xs_ref_vec[i][j];
      if( fabs(xs_ref_vec[i][j]) > min_value )
      {
        xs_dif /= xs_ref_vec[i][j];
      }

      fout << ene_vec[i][j] << "  "
           << xs_comp_vec[i][j] << "  " << xs_ref_vec[i][j] << "  " << xs_dif << endl;
    }
    fout.close();
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

int main(int argc, char *argv[])
{
  if( argc <= 1 )
  {
    throw runtime_error("Please set reference ace file name.");
  }
  else if( argc <= 2 )
  {
    throw runtime_error("Please set comparing ace file name.");
  }
  else if( argc <= 3 )
  {
    throw runtime_error("Please set output file name.");
  }


  string  ace_file_name_ref  = argv[1];
  string  ace_file_name_comp = argv[2];
  string  output_name        = argv[3];


  vector<vector<Integer> > nbt_xs_r, nbt_xs_c, int_xs_r, int_xs_c;
  vector<vector<Real> >    e_int_xs_r, e_int_xs_c, xs_tab_r, xs_tab_c;

  cout << "  read_ace_file (ref)  " << ace_file_name_ref  << endl;
  read_ace_file(ace_file_name_ref, e_int_xs_r, xs_tab_r);

  cout << "  read_ace_file (comp) " << ace_file_name_comp << endl;
  read_ace_file(ace_file_name_comp,  e_int_xs_c, xs_tab_c);


  vector<int>              ele_no;
  vector<vector<Integer> > nbt_xs_r_mod, nbt_xs_c_mod, int_xs_r_mod, int_xs_c_mod;
  vector<vector<Real> >    e_int_xs_r_mod, e_int_xs_c_mod, xs_tab_r_mod, xs_tab_c_mod;
  ele_no.resize(0);
  nbt_xs_r_mod.resize(0);
  int_xs_r_mod.resize(0);
  nbt_xs_c_mod.resize(0);
  int_xs_c_mod.resize(0);
  e_int_xs_r_mod.resize(0);
  xs_tab_r_mod.resize(0);
  e_int_xs_c_mod.resize(0);
  xs_tab_c_mod.resize(0);

  for(int i=0; i<react_no; i++)
  {
    int ene_no_r = static_cast<int>(e_int_xs_r[i].size());
    int ene_no_c = static_cast<int>(e_int_xs_c[i].size());
    if( ene_no_r > 0 && ene_no_c > 0 )
    {
      ele_no.push_back(i);
      e_int_xs_r_mod.push_back(e_int_xs_r[i]);
      e_int_xs_c_mod.push_back(e_int_xs_c[i]);
      xs_tab_r_mod.push_back(xs_tab_r[i]);
      xs_tab_c_mod.push_back(xs_tab_c[i]);

      vector<Integer> nbt_r, nbt_c, int_r, int_c;
      nbt_r.resize(1);
      nbt_c.resize(1);
      int_r.resize(1);
      int_c.resize(1);

      nbt_r[0] = int_lin_lin;
      nbt_c[0] = int_lin_lin;
      nbt_r[0] = ene_no_r;
      nbt_c[0] = ene_no_c;
      nbt_xs_r_mod.push_back(nbt_r);
      nbt_xs_c_mod.push_back(nbt_c);
      int_xs_r_mod.push_back(int_r);
      int_xs_c_mod.push_back(int_c);
    }
  }

  if( argc >= 5 )
  {
    string mg_structure_file_name = argv[4];
    cout << "  multi-group generation using " << mg_structure_file_name << endl;
    multi_group_generator(mg_structure_file_name, nbt_xs_r_mod, int_xs_r_mod,
                          e_int_xs_r_mod, xs_tab_r_mod);
    multi_group_generator(mg_structure_file_name, nbt_xs_c_mod, int_xs_c_mod,
                          e_int_xs_c_mod, xs_tab_c_mod);
  }
  else
  {
    cout << "  modify_xs_data" << endl;
    modify_xs_data(e_int_xs_r_mod, nbt_xs_c_mod, int_xs_c_mod,
                   e_int_xs_c_mod, xs_tab_c_mod);
  }

  for(int i=0; i<static_cast<int>(ele_no.size()); i++)
  {
    e_int_xs_r[ele_no[i]] = e_int_xs_r_mod[i];
    e_int_xs_r[ele_no[i]] = e_int_xs_r_mod[i];
    xs_tab_r[ele_no[i]]   = xs_tab_r_mod[i];
    xs_tab_r[ele_no[i]]   = xs_tab_r_mod[i];
  }

  //output ene, xs_r, xs_c
  cout << "  output_results" << endl;
  output_result(output_name, e_int_xs_r, xs_tab_r, xs_tab_c);
}

